import sys

# mergetb imports
import mergetb.mergetb as mergetb
from mergetb.workspace import User
from mergetb.types import StatusType
from mergetb.grpc_client import MergeGRPCError

# common test package imports
from .test_common import check_prompt_credentials

try:
    # Login to your account via the User object 
    username, password = check_prompt_credentials()
    u = User(username, password)
    u.login()

    user = u.get().user
    summary="""User info for %s:
    Name:          %s
    State:         %s
    Access Mode:   %s
    Uid:           %d
    Gid:           %d
    Projects:      %s
    Experiments:   %s
    Organizations: %s
    Facilities:    %s
    Admin:         %d
    Institution:   %s
    Category:      %s
    Email:         %s
    Country:       %s
    US State:      %s""" % (
        username,
        user.name,
        user.state,
        user.access_mode,
        user.uid,
        user.gid,
        user.projects,
        user.experiments,
        user.organizations,
        user.facilities,
        user.admin,
        user.institution,
        user.category,
        user.email,
        user.country,
        user.usstate,
    )
    print(summary)

    projects = mergetb.get_projects().projects

    print()
    print("User has the following projects:")
    for project in projects:
        summary="""Project: %s
    Description:  %s
    Members:      %s
    Experiments:  %s
    Access Mode:  %d
    GID:          %d
    Organization: %s
    Category:     %s
    Subcategory:  %s""" % (
            project.name,
            project.description,
            project.members,
            project.experiments,
            project.access_mode,
            project.gid,
            project.organization,
            project.category,
            project.subcategory,
        )
        print(summary)

# API errors are raised as MergeGRPCError
except MergeGRPCError as e:
    print(e)
except Exception as e:
    print(f"An unexpected error occurred: {e}")
else:
    print("test PASSED")
    sys.exit(0)

print("test FAILED")
sys.exit(1)



import asyncio
import paramiko

from mergetbapi.portal.v1 import Experiment as ExperimentV1
from mergetbapi.portal.v1 import *

from .mergetb import do_grpc_call
from .grpc_client import MergeGRPCClient
from .identity import Identity
from .model import Model
from .realize import Realization
from .materialize import Materialization
from .xdc import XDC

"""
We make the User object a sub-class of Identity so that methods of the latter can be invoked
by objects of the former

While Merge makes a distinction between these types, most users of this library likely will not
"""
class User(Identity):
    def __init__(self, username, password=None, channel=None, token=None):
        super().__init__(username, password, channel, token)

    async def _async_activate(self):
        return await WorkspaceStub(self.get_channel()).activate_user(
            ActivateUserRequest(username=self.username),
            metadata=self.get_auth_metadata()
        )

    async def _async_freeze(self):
        return await WorkspaceStub(self.get_channel()).freeze_user(
            FreezeUserRequest(username=self.username),
            metadata=self.get_auth_metadata()
        )

    async def _async_init(self):
        return await WorkspaceStub(self.get_channel()).init_user(
            InitUserRequest(username=self.username),
            metadata=self.get_auth_metadata()
        )

    async def _async_get(self):
        return await WorkspaceStub(self.get_channel()).get_user(
            GetUserRequest(
                username=self.username,
                status_ms=-1,
            ),
            metadata=self.get_auth_metadata()
        )

    async def _async_update(self, name=None, state=None, access_mode=None,
            organizations=None, projects=None, facilities=None, experiments=None,
            toggle_admin=False, institution=None, category=None, country=None,
            email=None, usstate=None):
        return await WorkspaceStub(self.get_channel()).update_user(
            UpdateUserRequest(
                username=self.username,
                name=name,
                state=state,
                access_mode=access_mode,
                organizations=organizations,
                projects=projects,
                facilities=facilities,
                experiments=experiments,
                toggle_admin=toggle_admin,
                institution=institution,
                category=category,
                country=country,
                email=email,
                usstate=usstate,
            ),
            metadata=self.get_auth_metadata()
        )

    async def _async_delete(self):
        return await WorkspaceStub(self.get_channel()).delete_user(
            DeleteUserRequest(user=self.username),
            metadata=self.get_auth_metadata()
        )

    async def _async_get_public_keys(self):
        return await WorkspaceStub(self.get_channel()).get_user_public_keys(
            GetUserPublicKeysRequest(user=self.username),
            metadata=self.get_auth_metadata()
        )

    async def _async_delete_public_keys(self):
        return await WorkspaceStub(self.get_channel()).delete_user_public_keys(
            DeleteUserPublicKeysRequest(user=self.username),
            metadata=self.get_auth_metadata()
        )

    async def _async_add_public_key(self, key):
        return await WorkspaceStub(self.get_channel()).add_user_public_key(
            AddUserPublicKeyRequest(user=self.username, key=key),
            metadata=self.get_auth_metadata()
        )

    async def _async_delete_public_key(self, key):
        return await WorkspaceStub(self.get_channel()).delete_user_public_key(
            DeleteUserPublicKeyRequest(user=self.username, fingerprint=key),
            metadata=self.get_auth_metadata()
        )

    def activate(self):
        return do_grpc_call(self._async_activate)

    def freeze(self):
        return do_grpc_call(self._async_freeze)

    def init(self):
        return do_grpc_call(self._async_init)

    def get(self):
        return do_grpc_call(self._async_get)

    def update(self, name=None, state=None, access_mode=None, organizations=None,
               projects=None, facilities=None, experiments=None, toggle_admin=False,
               institution=None, category=None, country=None, email=None, usstate=None):
        return do_grpc_call(self._async_update,
            name, state, access_mode, organizations, projects, facilities,
            experiments, toggle_admin, institution, category, country, email,
            usstate
        )

    def delete(self):
        return do_grpc_call(self._async_delete)

    def get_public_keys(self):
        return do_grpc_call(self._async_get_public_keys)

    def delete_public_keys(self):
        return do_grpc_call(self._async_delete_public_keys)

    def add_public_key(self, key):
        return do_grpc_call(self._async_add_public_key, key)

    def delete_public_key(self, key):
        return do_grpc_call(self._async_delete_public_key, key)

class Project(MergeGRPCClient):
    def __init__(self, name, channel=None, token=None):
        super().__init__(channel, token)
        self.name = name

    async def _async_create(self, username):
        return await WorkspaceStub(self.get_channel()).create_project(
            CreateProjectRequest(
                user=username, # strange that the API requires this ...
                project=self.name,
            ),
            metadata=self.get_auth_metadata()
        )

    async def _async_get(self):
        return await WorkspaceStub(self.get_channel()).get_project(
            GetProjectRequest(
                name=self.name,
                status_ms=-1,
            ),
            metadata=self.get_auth_metadata()
        )

    async def _async_update(self, description=None, access_mode=None, members=None, organization=None):
        return await WorkspaceStub(self.get_channel()).update_project(
            UpdateProjectRequest(
                name=self.name,
                description=description,
                access_mode=access_mode,
                members=members,
                organization=organization,
            ),
            metadata=self.get_auth_metadata()
        )

    async def _async_delete(self):
        return await WorkspaceStub(self.get_channel()).delete_project(
            DeleteProjectRequest(
                name=self.name,
            ),
            metadata=self.get_auth_metadata()
        )

    async def _async_get_members(self):
        return await WorkspaceStub(self.get_channel()).get_project_members(
            GetProjectMembersRequest(
                name=self.name,
            ),
            metadata=self.get_auth_metadata()
        )

    async def _async_add_member(self, username, member):
        return await WorkspaceStub(self.get_channel()).add_project_member(
            AddProjectMemberRequest(
                project=self.name,
                username=username,
                member=member,
            ),
            metadata=self.get_auth_metadata()
        )

    async def _async_get_member(self, username):
        return await WorkspaceStub(self.get_channel()).get_project_member(
            GetProjectMemberRequest(
                project=self.name,
                member=username,
            ),
            metadata=self.get_auth_metadata()
        )

    async def _async_update_member(self, username, member):
        return await WorkspaceStub(self.get_channel()).update_project_member(
            UpdateProjectMemberRequest(
                project=self.name,
                username=username,
                member=member,
            ),
            metadata=self.get_auth_metadata()
        )

    async def _async_delete_member(self, username):
        return await WorkspaceStub(self.get_channel()).delete_project_member(
            GetProjectMemberRequest(
                project=self.name,
                member=username,
            ),
            metadata=self.get_auth_metadata()
        )

    async def _async_get_experiments(self):
        return await WorkspaceStub(self.get_channel()).get_project_experiments(
            GetProjectExperimentsRequest(
                project=self.name,
            ),
            metadata=self.get_auth_metadata()
        )

    def create(self, username):
        return do_grpc_call(self._async_create, username)

    def get(self):
        return do_grpc_call(self._async_get)

    def update(self, description=None, access_mode=None, members=None, organization=None):
        return do_grpc_call(self._async_update,
            description, access_mode, members, organization
        )

    def delete(self):
        return do_grpc_call(self._async_delete)

    def add_member(self, username, member):
        return do_grpc_call(self._async_add_member, username, member)

    def get_member(self, username):
        return do_grpc_call(self._async_get_member, username)

    def update_member(self, username, member):
        return do_grpc_call(self._async_update_member, username, member)

    def delete_member(self, username):
        return do_grpc_call(self._async_delete_member, username)

    def get_members(self):
        return do_grpc_call(self._async_get_members)

    def get_experiments(self):
        return do_grpc_call(self._async_get_experiments)

class Organization(MergeGRPCClient):
    def __init__(self, name, channel=None, token=None):
        super().__init__(channel, token)
        self.name = name

    async def _async_create(self, username):
        return await WorkspaceStub(self.get_channel()).create_organization(
            CreateOrganizationRequest(
                user=username, # strange that the API requires this ...
                organization=self.name,
            ),
            metadata=self.get_auth_metadata()
        )

    async def _async_get(self):
        return await WorkspaceStub(self.get_channel()).get_organization(
            GetOrganizationRequest(
                name=self.name,
                status_ms=-1,
            ),
            metadata=self.get_auth_metadata()
        )

    async def _async_update(self, description=None, state=None, access_mode=None, members=None, projects=None):
        return await WorkspaceStub(self.get_channel()).update_organization(
            UpdateOrganizationRequest(
                name=self.name,
                description=description,
                state=state,
                access_mode=access_mode,
                members=members,
                projects=projects,
            ),
            metadata=self.get_auth_metadata()
        )

    async def _async_delete(self):
        return await WorkspaceStub(self.get_channel()).delete_organization(
            DeleteOrganizationRequest(
                name=self.name,
            ),
            metadata=self.get_auth_metadata()
        )

    async def _async_activate(self):
        return await WorkspaceStub(self.get_channel()).activate_organization(
            ActivateOrganizationRequest(
                organization=self.name,
            ),
            metadata=self.get_auth_metadata()
        )

    async def _async_freeze(self):
        return await WorkspaceStub(self.get_channel()).freeze_organization(
            FreezeOrganizationRequest(
                organization=self.name,
            ),
            metadata=self.get_auth_metadata()
        )

    async def _async_request_membership(self, username, member):
        return await WorkspaceStub(self.get_channel()).request_organization_membership(
            RequestOrganizationMembershipRequest(
                organization=self.name,
                id=username,
                kind=MembershipType.UserMember,
                member=member,
            ),
            metadata=self.get_auth_metadata()
        )

    async def _async_confirm_membership(self, username):
        return await WorkspaceStub(self.get_channel()).confirm_organization_membership(
            ConfirmOrganizationMembershipRequest(
                organization=self.name,
                id=username,
                kind=MembershipType.UserMember,
            ),
            metadata=self.get_auth_metadata()
        )

    async def _async_get_member(self, username):
        return await WorkspaceStub(self.get_channel()).get_organization_member(
            GetOrganizationMemberRequest(
                organization=self.name,
                username=username,
            ),
            metadata=self.get_auth_metadata()
        )

    async def _async_update_member(self, username, member):
        return await WorkspaceStub(self.get_channel()).update_organization_member(
            UpdateOrganizationMemberRequest(
                organization=self.name,
                username=username,
                member=member,
            ),
            metadata=self.get_auth_metadata()
        )

    async def _async_delete_member(self, username):
        return await WorkspaceStub(self.get_channel()).delete_organization_member(
            DeleteOrganizationMemberRequest(
                organization=self.name,
                username=username,
            ),
            metadata=self.get_auth_metadata()
        )

    async def _async_get_members(self):
        return await WorkspaceStub(self.get_channel()).get_organization_members(
            GetOrganizationMembersRequest(
                organization=self.name,
            ),
            metadata=self.get_auth_metadata()
        )

    async def _async_get_project(self, project):
        return await WorkspaceStub(self.get_channel()).get_organization_project(
            GetOrganizationProjectRequest(
                organization=self.name,
                project=project,
            ),
            metadata=self.get_auth_metadata()
        )

    async def _async_get_projects(self):
        return await WorkspaceStub(self.get_channel()).get_organization_projects(
            GetOrganizationProjectsRequest(
                organization=self.name,
            ),
            metadata=self.get_auth_metadata()
        )

    def create(self, username):
        return do_grpc_call(self._async_create, username)

    def get(self):
        return do_grpc_call(self._async_get)

    def update(self, description=None, state=None, access_mode=None, members=None, projects=None):
        return do_grpc_call(self._async_update,
            description, state, access_mode, members, projects,
        )

    def delete(self):
        return do_grpc_call(self._async_delete)

    def activate(self):
        return do_grpc_call(self._async_activate)

    def freeze(self):
        return do_grpc_call(self._async_freeze)

    def request_membership(self, username, member):
        return do_grpc_call(self._async_request_membership, username, member)

    def confirm_membership(self, username):
        return do_grpc_call(self._async_confirm_membership, username)

    def get_member(self, username):
        return do_grpc_call(self._async_get_member, username)

    def update_member(self, username, member):
        return do_grpc_call(self._async_update_member, username, member)

    def delete_member(self, username):
        return do_grpc_call(self._async_delete_member, username)

    def get_members(self):
        return do_grpc_call(self._async_get_members)

    def get_project(self, project):
        return do_grpc_call(self._async_get_project, project)

    def get_projects(self):
        return do_grpc_call(self._async_get_projects)

class Experiment(MergeGRPCClient):
    def __init__(self, name, project, description=None, channel=None, token=None):
        super().__init__(channel, token)
        self.name = name
        self.project = project
        self.description = description

    async def _async_create(self):
        return await WorkspaceStub(self.get_channel()).create_experiment(
            CreateExperimentRequest(
                ExperimentV1(
                    name=self.name,
                    project=self.project,
                    description=self.description,
                )
            ),
            metadata=self.get_auth_metadata()
        )

    async def _async_get(self, with_models=None):
        return await WorkspaceStub(self.get_channel()).get_experiment(
            GetExperimentRequest(
                experiment=self.name,
                project=self.project,
                with_models=with_models,
                status_ms=-1,
            ),
            metadata=self.get_auth_metadata()
        )

    async def _async_update(self, description=None, access_mode=None, creator=None, maintainers=None):
        return await WorkspaceStub(self.get_channel()).update_experiment(
            UpdateExperimentRequest(
                name=self.name,
                project=self.project,
                description=description,
                access_mode=access_mode,
                creator=creator,
                maintainers=maintainers,
            ),
            metadata=self.get_auth_metadata()
        )

    async def _async_delete(self):
        return await WorkspaceStub(self.get_channel()).delete_experiment(
            DeleteExperimentRequest(
                project=self.project,
                experiment=self.name,
            ),
            metadata=self.get_auth_metadata()
        )

    async def _async_get_revision(self, revision=None, encoding=None):
        return await WorkspaceStub(self.get_channel()).get_revision(
            GetRevisionRequest(
                name=self.name,
                project=self.project,
                revision=revision,
                encoding=encoding,
            ),
            metadata=self.get_auth_metadata()
        )

    async def _async_push_model(self, modelpath, branch=None, tag=None):
        return await ModelStub(self.get_channel()).push(
            PushRequest(
                project=self.project,
                experiment=self.name,
                model=Model(modelpath).contents(),
                branch=branch,
                tag=tag,
            ),
            metadata=self.get_auth_metadata()
        )

    def create(self):
        return do_grpc_call(self._async_create)

    def get(self, with_models=None):
        return do_grpc_call(self._async_get, with_models)

    def update(self, description=None, access_mode=None, creator=None, maintainers=None):
        return do_grpc_call(self._async_update, description, access_mode, creator, maintainers)

    def delete(self):
        return do_grpc_call(self._async_delete)

    def get_revision(self, revision=None, encoding=None):
        return do_grpc_call(self._async_update, revision, encoding)

    def push_model(self, modelpath, branch=None, tag=None):
        return do_grpc_call(self._async_push_model, modelpath, branch, tag)

    def realize(self, realization, revision=None, tag=None, branch=None, duration=None):
        # call into the realize package to create an object and realize it
        return Realization(
            realization, self.name, self.project,
            revision=revision, tag=tag, branch=branch,
            channel=self.channel,
        ).realize()

    def relinquish(self, realization):
        # call into the realize package to relinquish
        return Realization(
            realization, self.name, self.project,
        ).relinquish()

    def materialize(self, realization):
        # call into the materialize package to create an object and materialize it
        return Materialization(
            realization, self.name, self.project,
            channel=self.channel,
        ).materialize()

    def dematerialize(self, realization):
        # call into the materialize package to dematerialize
        return Materialization(
            realization, self.name, self.project,
            channel=self.channel,
        ).dematerialize()

    # SPHERE trappings - alias for realize/relinquish
    def reserve(self, realization, revision=None, tag=None, branch=None, duration=None):
        return self.realize(realization, revision, tag, branch, duration)

    def free(self, realization):
        return self.relinquish(realization)

    # SPHERE trappings - alias for materialize/dematerialize
    def activate(self, realization):
        return self.materialize(realization)

    def deactivate(self, realization):
        return self.dematerialize(realization)

    def attach_xdc(self, realization, xdcname, xdcproject):
        # call into the XDC package to attach
        return XDC(
            xdcname, xdcproject,
            channel=self.channel,
        ).attach(realization, self.name, self.project)

    # Attach the experiment to xdc
    def detach_xdc(self, xdcname, xdcproject):
        # call into the XDC package to detach
        return XDC(
            xdcname, xdcproject,
            channel=self.channel,
        ).detach()

    # Execute command on node using SSH
    def exec_on_node(self, username, node, cmd):
        ssh_client = paramiko.SSHClient()
        private_key = paramiko.RSAKey.from_private_key_file("/home/"+username+"/.ssh/merge_key")
        ssh_client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        ssh_client.connect(node, username=username, pkey=private_key)
        (stdin, stdout, stderr) = ssh_client.exec_command(cmd)
        lines = stdout.readlines()
        elines = stderr.readlines()
        ssh_client.close()
        return (lines, elines)



import asyncio
import tempfile
import validators
from urllib.request import urlretrieve

from mergetbapi.portal.v1 import *

from .mergetb import do_grpc_call
from .grpc_client import MergeGRPCClient

class Model(MergeGRPCClient):
    def __init__(self, modelpath, channel=None, token=None):
        super().__init__(channel, token)
        self.model = modelpath

    async def _async_compile(self):
        return await ModelStub(self.get_channel()).compile(
            CompileRequest(model=self.contents()),
            metadata=self.get_auth_metadata()
        )
 
    def contents(self):
        if validators.url(self.model):
            with tempfile.NamedTemporaryFile() as tmp:
                urlretrieve(self.model, tmp.name)
                return open(tmp.name, 'r').read()
        else:
            return open(self.model, 'r').read()

    def compile(self):
        return do_grpc_call(self._async_compile)



import sys
from timeout_timer import timeout, TimeoutInterrupt 

# mergetb imports
import mergetb.mergetb as mergetb
from mergetb.identity import Identity
from mergetb.workspace import Experiment
from mergetb.realize import Realization
from mergetb.materialize import Materialization
from mergetb.types import StatusType
from mergetb.grpc_client import MergeGRPCError

# common test package imports
from .test_common import check_prompt_credentials

# Wait 15 seconds for a realization to complete
RLZ_TIMEOUT = 15

# Wait 60 seconds for a materialization to complete
MTZ_TIMEOUT = 60

exp = None
try:
    # Login to your account
    username, password = check_prompt_credentials()
    id = Identity(username, password)
    id.login()
    print("Logged in for user %s" % username)

    # Create experiment
    exp = Experiment("mergetbtest", username)
    exp.create()

    # Push a model and tag it v1
    exp.push_model("tests/data/model_test.py", tag="v1")

    # Realize tag v1
    rlz = Realization("v1", exp.name, exp.project, tag="v1")
    rlz.realize()

    # wait up to 10 seconds for the realization to succeed
    try:
        with timeout(RLZ_TIMEOUT):
            status = StatusType.Undefined 

            while True:
                resp = rlz.get()
                status = resp.status.highest_status

                if status == StatusType.Success:
                    print("realization succeeded")
                    break
                elif status == StatusType.Deleted:
                    raise Exception("realization deleted")
                elif status == StatusType.Pending:
                    print("realization pending")
                elif status == StatusType.Unresponsive:
                    raise Exception("realization unresponsive")
                elif status == StatusType.Processing:
                    print("realization processing")
                elif status == StatusType.Error:
                    raise Exception("realization error")
                else:
                    raise Exception("unknown realization error")
    except TimeoutInterrupt:
        raise Exception("realization did not succeed within %d seconds" % RLZ_TIMEOUT)

    resp = rlz.get()
    print("Realization expires at %s" % (resp.result.realization.expires))

    # set the rlz expiration time to 2 weeks from now
    rlz.update("2w")

    # get the new realization and check the expiration
    resp = rlz.get()
    print("Realization expires at %s" % (resp.result.realization.expires))

    # materialize it
    mtz = Materialization("v1", exp.name, exp.project)
    mtz.materialize()

    try:
        with timeout(MTZ_TIMEOUT):
            status = StatusType.Undefined 

            while True:
                resp = mtz.get()
                status = resp.status.highest_status

                if status == StatusType.Success:
                    print("materialization succeeded")
                    break
                elif status == StatusType.Deleted:
                    raise Exception("materialization deleted")
                elif status == StatusType.Pending:
                    print("materialization pending")
                elif status == StatusType.Unresponsive:
                    raise Exception("materialization unresponsive")
                elif status == StatusType.Processing:
                    print("materialization processing")
                elif status == StatusType.Error:
                    print("materialization error")
                    #raise Exception("materialization error")
                else:
                    raise Exception("unknown materialization error")
    except TimeoutInterrupt:
        raise Exception("materialization did not succeed within %d seconds" % MTZ_TIMEOUT)

# API errors are raised as MergeGRPCError
except MergeGRPCError as e:
    print(e)
except Exception as e:
    print(f"An unexpected error occurred: {e}")
else:
    print("test PASSED")
    sys.exit(0)
finally:
    # teardown the test
    if exp is not None:
        exp.delete()

print("test FAILED")
sys.exit(1)



from grpclib.client import Channel
from google.rpc.error_details_pb2 import BadRequest

DEFAULT_PORTAL_URL = "grpc.sphere-testbed.net"
DEFAULT_PORTAL_PORT = 443
DEFAULT_PORTAL_SSL = True

class MergeGRPCConfig:
    def __init__(self, server, port=DEFAULT_PORTAL_PORT, ssl=DEFAULT_PORTAL_SSL):
        self.server = server
        self.port = port
        self.ssl = ssl

    def __str__(self):
        return "%s:%d (ssl=%d)" % (
            self.server, self.port, self.ssl
        )

dflt_grpc_config  = MergeGRPCConfig(DEFAULT_PORTAL_URL)
dflt_grpc_token   = None
dflt_grpc_channel = None

# initialize a GRPC channel
def _initialize_grpc_channel(grpc_config):
    return Channel(
        grpc_config.server,
        grpc_config.port,
        ssl=grpc_config.ssl,
    )

def set_grpc_config(config):
    global dflt_grpc_config
    dflt_grpc_config = config

class MergeGRPCClient:
    def __init__(self, channel=None, token=None):
        self.token = token
        self.channel = channel
        self.inited_channel = False

    def __str__(self):
        return self.token if self.token is not None else "(NULL)"

    def __del__(self):
        if self.inited_channel:
            self.channel.close()
            self.inited_channel = False

    def get_channel(self):
        if self.channel is  None:
            global dflt_grpc_config
            self.channel = _initialize_grpc_channel(dflt_grpc_config)
            self.inited_channel = True

        return self.channel

    def set_bearer_token(self, token):
        self.token = token

        # save in the default global token
        global dflt_grpc_token
        dflt_grpc_token = token

    def get_bearer_token(self):
        global dflt_grpc_token
        return self.token if self.token is not None else dflt_grpc_token

    def get_auth_metadata(self):
        return {
            "authorization": "Bearer %s" % self.get_bearer_token(),
        }

class MergeGRPCError(Exception):
    def __init__(self, grpc_error):
        self.grpc_error = grpc_error
        super().__init__(self.grpc_error.message)

    def __str__(self):
        # status: grpclib.const.Status
        # message: str
        # details: any
        if self.grpc_error.details is not None:
            for d in self.grpc_error.details:
                if isinstance(d, BadRequest):
                    return str(MergeError(d))

        return """
An Uncategorized MergeTB API Error Occurred:
    Message:     %s
    Status Code: %d (%s)
    Details:     %s""" % ( 
        self.grpc_error.message, 
        self.grpc_error.status.value, 
        self.grpc_error.status.name, 
        self.grpc_error.details
    )

    def __repr__(self):
        return self.__str__()

class MergeError():
    def __init__(self, badrequest):
        self.fields = {}
        for v in badrequest.field_violations:
            self.fields[v.field.lower()] = v.description

    def __str__(self):
        title = self.get_field('title')
        detail = self.get_field('detail')
        instance = self.get_field('instance')
        evidence = self.get_field('evidence')
        type = self.get_field('type')
        timestamp = self.get_field('timestamp')
        
        s = '\nA MergeTB API Error Occurred:'
        if title != '':
            s = s + '\n    Title:      %s' % title
        if detail != '':
            s = s + '\n    Detail:     %s' % detail
        if instance != '':
            s = s + '\n    Instance:   %s' % instance
        if evidence != '':
            s = s + '\n    Evidence:   %s' % evidence
        if type != '':
            s = s + '\n    Type:       %s' % type
        if timestamp != '':
            s = s + '\n    Timestamp:  %s' % timestamp

        return s

    def __repr__(self):
        return self.__str__()

    def get_field(self, field):
        return self.fields[field] if field in self.fields else ''



import asyncio

from mergetbapi.portal.v1 import *

from .mergetb import do_grpc_call
from .grpc_client import MergeGRPCClient

class Materialization(MergeGRPCClient):
    def __init__(self, name, experiment, project, channel=None, token=None):
        super().__init__(channel, token)
        self.name = name
        self.experiment = experiment
        self.project = project

    async def _async_materialize(self):
        return await MaterializeStub(self.get_channel()).materialize(
            MaterializeRequest(
                project=self.project,
                experiment=self.experiment,
                realization=self.name,
            ), metadata=self.get_auth_metadata()
        )

    async def _async_get(self):
        return await MaterializeStub(self.get_channel()).get_materialization_v2(
            GetMaterializationRequestV2(
                project=self.project,
                experiment=self.experiment,
                realization=self.name,
                status_ms=-1,
            ), metadata=self.get_auth_metadata()
        )

    async def _async_dematerialize(self):
        return await MaterializeStub(self.get_channel()).dematerialize(
            DematerializeRequest(
                project=self.project,
                experiment=self.experiment,
                realization=self.name,
            ), metadata=self.get_auth_metadata()
        )

    def materialize(self):
        return do_grpc_call(self._async_materialize)

    def get(self):
        return do_grpc_call(self._async_get)

    def dematerialize(self):
        return do_grpc_call(self._async_dematerialize)

    # SPHERE trappings - alias for materialize/dematerialize
    def activate(self):
        return self.materialize()

    def deactivate(self):
        return self.dematerialize()



import asyncio

from mergetbapi.portal.v1 import *
from .mergetb import do_grpc_call
from .grpc_client import MergeGRPCClient

class Realization(MergeGRPCClient):
    def __init__(self, name, experiment, project, revision=None, tag=None, branch=None,
                 duration=None, channel=None, token=None):

        super().__init__(channel, token)
        self.name = name
        self.experiment = experiment
        self.project = project
        self.revision = revision
        self.tag = tag
        self.branch = branch
        self.duration = duration

    async def _async_realize(self):
        return await RealizeStub(self.get_channel()).realize(
            RealizeRequest(
                project=self.project,
                experiment=self.experiment,
                realization=self.name,
                revision=self.revision,
                tag=self.tag,
                branch=self.branch,
                duration=self.duration,
            ), metadata=self.get_auth_metadata()
        )

    async def _async_get(self):
        return await RealizeStub(self.get_channel()).get_realization(
            GetRealizationRequest(
                project=self.project,
                experiment=self.experiment,
                realization=self.name,
                status_ms=-1,
            ), metadata=self.get_auth_metadata()
        )

    async def _async_update(self, duration=None):
        self.duration = duration

        return await RealizeStub(self.get_channel()).update_realization(
            UpdateRealizationRequest(
                project=self.project,
                experiment=self.experiment,
                realization=self.name,
                duration=ReservationDuration(
                    when=ReservationDurationCode.given,
                    duration=self.duration,
                )
            ), metadata=self.get_auth_metadata()
        )

    async def _async_relinquish(self):
        return await RealizeStub(self.get_channel()).relinquish(
            RelinquishRequest(
                project=self.project,
                experiment=self.experiment,
                realization=self.name,
            ), metadata=self.get_auth_metadata()
        )

    def realize(self):
        return do_grpc_call(self._async_realize)

    def get(self):
        return do_grpc_call(self._async_get)

    def update(self, duration=None):
        return do_grpc_call(self._async_update, duration)

    def relinquish(self):
        return do_grpc_call(self._async_relinquish)

    # SPHERE trappings; alias for realize/relinquish
    def reserve(self):
        return self.realize()

    def free(self):
        return self.relinquish()



import sys

# mergetb imports
from mergetb.identity import Identity
from mergetb.grpc_client import MergeGRPCError

# common test package imports
from .test_common import check_prompt_credentials

try:
    # Login to your account
    username, password = check_prompt_credentials()
    id = Identity(username, password)
    id.login()
    print("Logged in for user %s" % username)

    # Admin only
    resp = id.get().identity
    summary="""User info for %s:
    Email:  %s
    Admin:  %s
    Traits: %s""" % (
        username,
        resp.email,
        resp.admin,
        resp.traits,
    )
    print(summary)

# API errors are raised as MergeGRPCError
except MergeGRPCError as e:
    print(e)
except Exception as e:
    print(f"An unexpected error occurred: {e}")
else:
    print("test PASSED")
    sys.exit(0)

print("test FAILED")
sys.exit(1)



import asyncio

from mergetbapi.portal.v1 import *

from .mergetb import do_grpc_call
from .grpc_client import MergeGRPCClient

class XDC(MergeGRPCClient):
    def __init__(self, name, project, xdctype=XdcType.personal, channel=None, token=None):
        super().__init__(channel, token)
        self.name = name
        self.project = project
        self.xdctype = xdctype

    async def _async_create(self):
        return await XdcStub(self.get_channel()).create_xdc(
            CreateXdcRequest(
                project=self.project,
                xdc=self.name,
                type=self.xdctype,
            ), metadata=self.get_auth_metadata()
        )

    async def _async_get(self):
        return await XdcStub(self.get_channel()).get_xdc(
            GetXdcRequest(
                project=self.project,
                xdc=self.name,
                status_ms=-1,
            ), metadata=self.get_auth_metadata()
        )

    async def _async_delete(self):
        return await XdcStub(self.get_channel()).delete_xdc(
            DeleteXdcRequest(
                project=self.project,
                xdc=self.name,
            ), metadata=self.get_auth_metadata()
        )

    async def _async_attach(self, realization, experiment, project):
        return await XdcStub(self.get_channel()).attach_xdc(
            AttachXdcRequest(
                xdc=self.name,
                project=self.project,
                experiment=experiment,
                realization=realization,
                realization_project=project,
            ), metadata=self.get_auth_metadata()
        )

    async def _async_detach(self):
        return await XdcStub(self.get_channel()).detach_xdc(
            DetachXdcRequest(
                xdc=self.name,
                project=self.project,
            ), metadata=self.get_auth_metadata()
        )

    async def _async_detach(self):
        return await XdcStub(self.get_channel()).detach_xdc(
            DetachXdcRequest(
                xdc=self.name,
                project=self.project,
            ), metadata=self.get_auth_metadata()
        )

    def create(self):
        return do_grpc_call(self._async_create)

    def get(self):
        return do_grpc_call(self._async_get)

    def delete(self):
        return do_grpc_call(self._async_delete)

    def attach(self, realization, experiment, project):
        return do_grpc_call(self._async_attach, realization, experiment, project)

    def detach(self):
        return do_grpc_call(self._async_detach)



import asyncio
import threading
from typing import Awaitable, TypeVar

from grpclib.exceptions import GRPCError, StreamTerminatedError
from grpclib.client import Channel

from mergetbapi.portal.v1 import *
from .grpc_client import MergeGRPCClient, MergeGRPCError

"""
<<<<<<<< Cribbed from: https://stackoverflow.com/a/69514930
"""
T = TypeVar("T")
def _start_background_loop(loop):
    asyncio.set_event_loop(loop)
    loop.run_forever()

_LOOP = asyncio.new_event_loop()
_LOOP_THREAD = threading.Thread(
    target=_start_background_loop, args=(_LOOP,), daemon=True
)
_LOOP_THREAD.start()

def asyncio_run(coro: Awaitable[T], timeout=30) -> T:
    """
    Runs the coroutine in an event loop running on a background thread,
    and blocks the current thread until it returns a result.
    This plays well with gevent, since it can yield on the Future result call.

    :param coro: A coroutine, typically an async method
    :param timeout: How many seconds we should wait for a result before raising an error
    """
    return asyncio.run_coroutine_threadsafe(coro, _LOOP).result(timeout=timeout)


def asyncio_gather(*futures, return_exceptions=False) -> list:
    """
    A version of asyncio.gather that runs on the internal event loop
    """
    async def gather():
        return await asyncio.gather(*futures, return_exceptions=return_exceptions)

    return asyncio.run_coroutine_threadsafe(gather(), loop=_LOOP).result()
"""
>>>>>>>>
"""

# initialize a default shared grpc_client for functions in this package
dflt_grpc_client = MergeGRPCClient()

async def _async_get_identities(grpc_client=dflt_grpc_client):
    return await IdentityStub(grpc_client.get_channel()).list_identities(
        ListIdentityRequest(),
        metadata=grpc_client.get_auth_metadata(),
    )

async def _async_get_users(filter=None, grpc_client=dflt_grpc_client):
    return await WorkspaceStub(grpc_client.get_channel()).get_users(
        GetUsersRequest(
            filter=filter
        ),
        metadata=grpc_client.get_auth_metadata()
    )

async def _async_get_experiments(filter=None, grpc_client=dflt_grpc_client):
    return await WorkspaceStub(grpc_client.get_channel()).get_experiments(
        GetExperimentsRequest(
            filter=filter,
        ),
        metadata=grpc_client.get_auth_metadata()
    )

async def _async_get_projects(filter=None, grpc_client=dflt_grpc_client):
    return await WorkspaceStub(grpc_client.get_channel()).get_projects(
        GetProjectsRequest(
            filter=filter,
        ),
        metadata=grpc_client.get_auth_metadata()
    )

async def _async_get_organizations(filter=None, grpc_client=dflt_grpc_client):
    return await WorkspaceStub(grpc_client.get_channel()).get_organizations(
        GetOrganizationsRequest(
            filter=filter,
        ),
        metadata=grpc_client.get_auth_metadata()
    )

async def _async_get_realizations(filter=None, grpc_client=dflt_grpc_client):
    return await RealizeStub(grpc_client.get_channel()).get_realizations(
        GetRealizationsRequest(
            filter=filter,
        ),
        metadata=grpc_client.get_auth_metadata()
    )

async def _async_get_materializations(filter=None, grpc_client=dflt_grpc_client):
    return await MaterializeStub(grpc_client.get_channel()).get_materializations_v2(
        GetMaterializationsRequestV2(
            filter=filter,
        ),
        metadata=grpc_client.get_auth_metadata()
    )

async def _async_get_xdcs(project=None, grpc_client=dflt_grpc_client):
    return await XdcStub(grpc_client.get_channel()).list_xd_cs(
        ListXdCsRequest(
            project=project,
        ),
        metadata=grpc_client.get_auth_metadata()
    )

async def _async_get_xdc_jump_hosts(grpc_client=dflt_grpc_client):
    return await XdcStub(grpc_client.get_channel()).get_xdc_jump_hosts(
        GetXdcJumpHostsRequest(),
        metadata=grpc_client.get_auth_metadata()
    )

async def _async_get_entity_type_configurations(grpc_client=dflt_grpc_client):
    return await WorkspaceStub(grpc_client.get_channel()).get_entity_type_configurations(
        GetEntityTypeConfigurationsRequest(),
        metadata=grpc_client.get_auth_metadata()
    )

async def _async_update_entity_type_configurations(types=None, patch_strategy=None, grpc_client=dflt_grpc_client):
    return await WorkspaceStub(grpc_client.get_channel()).update_entity_type_configurations(
        UpdateEntityTypeConfigurationsRequest(
            types=types,
            patch_strategy=patch_strategy,
        ),
        metadata=grpc_client.get_auth_metadata()
    )

async def _async_get_user_configurations(grpc_client=dflt_grpc_client):
    return await WorkspaceStub(grpc_client.get_channel()).get_user_configurations(
        GetUserConfigurationsRequest(),
        metadata=grpc_client.get_auth_metadata()
    )

async def _async_update_user_configurations(institutions=None, categories=None, patch_strategy=None, grpc_client=dflt_grpc_client):
    return await WorkspaceStub(grpc_client.get_channel()).update_user_configurations(
        UpdateUserConfigurationsRequest(
            institutions=institutions,
            categories=categories,
            patch_strategy=patch_strategy,
        ),
        metadata=grpc_client.get_auth_metadata()
    )

def get_identities(grpc_client=dflt_grpc_client):
    return do_grpc_call(_async_get_identities, grpc_client)

def get_users(filter=None, grpc_client=dflt_grpc_client):
    return do_grpc_call(_async_get_users, filter, grpc_client)

def get_experiments(filter=None, grpc_client=dflt_grpc_client):
    return do_grpc_call(_async_get_experiments, filter, grpc_client)

def get_projects(filter=None, grpc_client=dflt_grpc_client):
    return do_grpc_call(_async_get_projects, filter, grpc_client)

def get_organizations(filter=None, grpc_client=dflt_grpc_client):
    return do_grpc_call(_async_get_organizations, filter, grpc_client)

def get_realizations(filter=None, grpc_client=dflt_grpc_client):
    return do_grpc_call(_async_get_realizations, filter, grpc_client)

def get_materializations(filter=None, grpc_client=dflt_grpc_client):
    return do_grpc_call(_async_get_materializations, filter, grpc_client)

def get_xdcs(project=None, grpc_client=dflt_grpc_client):
    return do_grpc_call(_async_get_xdcs, project, grpc_client)

def get_xdc_jump_hosts(grpc_client=dflt_grpc_client):
    return do_grpc_call(_async_get_xdc_jump_hosts, grpc_client)

def get_entity_type_configurations(grpc_client=dflt_grpc_client):
    return do_grpc_call(_async_get_entity_type_configurations, grpc_client)

def update_entity_type_configurations(types=None, patch_strategy=None, grpc_client=dflt_grpc_client):
    return do_grpc_call(_async_update_entity_type_configurations,
        types, patch_strategy, grpc_client
    )

def get_user_configurations(grpc_client=dflt_grpc_client):
    return do_grpc_call(_async_get_user_configurations, grpc_client)

def update_user_configurations(institutions=None, categories=None, patch_strategy=None, grpc_client=dflt_grpc_client):
    return do_grpc_call(_async_update_entity_type_configurations,
        institutions, categories, patch_strategy, grpc_client
    )

# execute an asynchronous MergeTB API call and handle exceptions that can happen
GRPC_CALL_TRIES=5
def do_grpc_call(fn, *args, **kwargs):
    attempt = 1
    while True:
        try:
            return asyncio_run(fn(*args, **kwargs))
        except GRPCError as grpc_error:
            raise MergeGRPCError(grpc_error)
        except StreamTerminatedError as e:
            # grpclib does not process GOAWAY frames properly
            # These tend to indicate transient problems, so we can generally
            # retry the call and it will succeed.

            attempt = attempt + 1
            if attempt > GRPC_CALL_TRIES:
                raise e

            print("  [stream_terminated: attempting try %d out of %d]" %
                (attempt, GRPC_CALL_TRIES)
            )



import asyncio

from mergetbapi.portal.v1 import *

from .grpc_client import MergeGRPCClient
from .mergetb import do_grpc_call

class Identity(MergeGRPCClient):
    def __init__(self, username, password=None, channel=None, token=None):
        super().__init__(channel, token)
        self.username = username
        self.password = password

    async def _async_login(self):
        response = await IdentityStub(self.get_channel()).login(
            LoginRequest(username=self.username, password=self.password)
        )
        self.set_bearer_token(response.token)
        return response

    async def _async_logout(self):
        return await IdentityStub(self.get_channel()).logout(
            LogoutRequest(username=self.username),
            metadata=self.get_auth_metadata(),
        )

    async def _async_get(self):
        return await IdentityStub(self.get_channel()).get_identity(
            GetIdentityRequest(username=self.username),
            metadata=self.get_auth_metadata(),
        )

    async def _async_register(self, password, email, institution, category, country, usstate, name, admin=False):
        return await IdentityStub(self.get_channel()).register(RegisterRequest(
            username=self.username,
            email=email,
            password=password,
            institution=institution,
            category=category,
            country=country,
            usstate=usstate,
            name=name,
            admin=admin
        ))

    async def _async_unregister(self):
        return IdentityStub(self.get_channel()).register(UnregisterRequest(
            username=self.username,
            metadata=self.get_auth_metadata(),
        ))

    def login(self):
        return do_grpc_call(self._async_login)

    def logout(self):
        return do_grpc_call(self._async_logout)

    def get(self):
        return do_grpc_call(self._async_get)

    def register(self, password, email, institution, category, country, usstate, name, admin=False):
        return do_grpc_call(self._async_register,
            email, institution, category, country, usstate, name, admin
        )

    def unregister(self):
        return do_grpc_call(self._async_unregister)



#!/usr/bin/env python

import os
import io
import sys
from setuptools import setup

try:
    VERSION=os.environ['VERSION']
except KeyError:
    print('Must provide VERSION as an environment variable')
    sys.exit(1)

APIVERSION='v1.3.9'
try:
    APIVERSION=os.environ['APIVERSION']
except KeyError:
    print('No APIVERSION provided; using %s' % APIVERSION)

print('building package mergetb==%s, depends on mergetbapi>=%s' % (
    VERSION, APIVERSION
))

NAME='mergetb'
DESCRIPTION='MergeTB Python client library',
URL = 'https://gitlab.com/mergetb/portal/python-client'
EMAIL = 'bkocolos@isi.edu'
AUTHOR = 'Brian Kocoloski'
REQUIRES_PYTHON = '>=3'

REQUIRED = [
    'betterproto>=2.0.0b6',
    'grpclib>=0.4.7',
    'paramiko>=3.4.0',
    'validators>=0.33.0',
    'timeout-timer>=0.2.0',
    'pwinput>=1.0.3',
    'googleapis-common-protos>=1.63.0',
    'mergetbapi>=%s' % APIVERSION
]
CLASSIFIERS = [
    'Programming Language :: Python :: 3',
    'License :: OSI Approved :: MIT License',
    'Operating System :: OS Independent',
    'Intended Audience :: Developers',
    'Intended Audience :: End Users/Desktop',
    'Development Status :: 4 - Beta',
]

HERE = os.path.abspath(os.path.dirname(__file__))
with io.open(os.path.join(HERE, 'README.md')) as f:
    LDESCRIPTION = f.read()

setup(
    name=NAME,
    version=VERSION,
    author=AUTHOR,
    author_email=EMAIL,
    python_requires=REQUIRES_PYTHON,
    url=URL,
    description=DESCRIPTION,
    long_description=LDESCRIPTION,
    long_description_content_type='text/markdown',
    license='MIT',
    packages=['mergetb'],
    install_requires=REQUIRED,
    include_package_data=True,
    zip_safe=False,
    classifiers=CLASSIFIERS,
)
